(define (domain rpg)

    (:requirements :strips :typing :negative-preconditions)

    (:types
        square
        hero
    )

    (:predicates
        (hero_at ?hero - hero ?x - square)
        (monster_awake ?x - square)
        (trap_armed ?x - square)
        (sword_at ?x - square)
        (visited ?x - square)
        (valid_move ?from - square ?to - square)
        (empty_handed ?x - hero)
    )

    (:action move
        :parameters (?hero - hero ?from - square ?to - square)
        :precondition (and
            (hero_at ?hero ?from) ; Hero must be at ?from square.
            (valid_move ?from ?to) ; It must to be a valid move on the board.
            (not (visited ?from)) ; The square can not have been visited.
            (not (visited ?to)) ; The square can not have been visited.
            (not (trap_armed ?from)) ; Hero can not move with an armed trap.
            (not (monster_awake ?from)) ; Hero can not move with the monster awake.
            (not (and
                (monster_awake ?to)
                (empty_handed ?hero)
                ; Condition to ensure that the hero does not go to ?to, which has a monster, empty-handed.
            ))
        )
        
        :effect (and
            (hero_at ?hero ?to) ; Hero changes position.
            (visited ?from) ; ?from has now been visited.
        )
    )

    (:action get_sword
        :parameters (?hero - hero ?location - square)
        :precondition (and
            (hero_at ?hero ?location) ; Hero must be at ?location square.
            (not (visited ?location)) ; The square can not have been visited.
            (sword_at ?location) ; Sword must be at ?location square.
            (empty_handed ?hero) ; Hero does not have a sword.
        )
        
        :effect (and
            (not (empty_handed ?hero)) ; Hero is no longer empty-handed.
        )
    )

    (:action drop_sword
        :parameters (?hero - hero ?location - square)
        :precondition (and
            (hero_at ?hero ?location) ; Hero must be at ?location square.
            (trap_armed ?location) ; Trap is armed at ?location.
        )
        
        :effect (and
            (empty_handed ?hero) ; Hero no longer holds a sword.
        )
    )

    (:action fall_asleep_monster
        :parameters (?hero - hero ?location - square)
        :precondition (and
            (hero_at ?hero ?location) ; Hero must be at ?location square.
            (monster_awake ?location) ; Monster must be awake at ?location square.
            (not (empty_handed ?hero)) ; Hero have the sword.
        )
        
        :effect (and
            (not (monster_awake ?location)) ; Fall asleep the monster. Imagine that the sword has some kind of sleeping medicine.
            ; I could move the hero forward here, but I think it is more consistent to let the move action do that.
        )
    )

    (:action disarm_trap
        :parameters (?hero - hero ?location - square)
        :precondition (and
            (hero_at ?hero ?location) ; Hero must be at ?location square.
            (trap_armed ?location) ; Trap is armed at ?location.
            (empty_handed ?hero) ; Hero can not have a sword.
        )
        
        :effect (and
            (not (trap_armed ?location)) ; Disarm the trap at ?location.
        )
    )
)